__version__ = "0.1.28"
__author__ = "FinOL Contributors"

#
"""
python setup.py sdist build
twine upload dist/*

pip install --upgrade finol
python -m finol.FinOLAPP
python -m finol.main.main.py
"""

